#ifndef __FORKJOIN_HPP__
#define __FORKJOIN_HPP__

#include <atomic>
#include <cstdint>
#include <limits>
#include <memory>
#include <optional>
#include <utility>

#include <coroutine.hpp>
#include <result.hpp>

#include <riften/forkpool.hpp>


namespace typon
{

  struct coroutine_node
  {
    coroutine_node * _next;
    coroutine_handle<> _coroutine;
  };


  struct coroutine_list
  {
    coroutine_node * _first = nullptr;

    template <typename Promise>
    void insert(coroutine_handle<Promise> coroutine) noexcept
    {
      coroutine.promise()._node = { _first, coroutine };
      _first = &(coroutine.promise()._node);
    }

    void clear() noexcept
    {
      auto next = _first;
      while(next)
      {
        auto node = next;
        next = next->_next;
        node->_coroutine.destroy();
      }
      _first = nullptr;
    }
  };


  struct continuation_handle
  {
    coroutine_handle<> _coroutine;
    coroutine_handle<> * _continuation;

    continuation_handle() noexcept {}

    template <typename Promise>
    continuation_handle(coroutine_handle<Promise> coroutine) noexcept
      : _coroutine(coroutine), _continuation(&(coroutine.promise()._continuation))
    {}

    continuation_handle(noop_coroutine_handle coroutine) noexcept
      : _coroutine(coroutine), _continuation(nullptr)
    {}

    operator coroutine_handle<>() noexcept
    {
      return _coroutine;
    }

    coroutine_handle<> continuation() noexcept
    {
      if (_coroutine.done())
      {
        return *_continuation;
      }
      return _coroutine;
    }
  };

    template <typename Promise, bool = false>
    struct Future
    {
      coroutine_handle<Promise> _coroutine;
      bool _owning;

      Future(coroutine_handle<Promise> coroutine, bool owning)
        : _coroutine(coroutine), _owning(owning)
      {}

      Future(const Future &) = delete;

      Future& operator=(const Future &) = delete;

      Future(Future && other) noexcept
      {
        *this = other;
        other.owning = false;
      }

      Future& operator=(Future && other) noexcept
      {
        *this = other;
        other._owning = false;
      }

      ~Future()
      {
        if (_owning)
        {
          _coroutine.destroy();
        }
      }

      auto get()
      {
        return _coroutine.promise().get();
      }
    };


  template <typename Promise>
  struct Future<Promise, true>
  {
    using result_type = Result<typename Promise::value_type>;
    union
    {
      coroutine_handle<Promise> _coroutine;
      result_type _result;
    };
    bool _owning;

    Future(coroutine_handle<Promise> coroutine, bool owning)
    {
      _owning = owning;
      if (owning)
      {
        std::construct_at(&(_result), *static_cast<result_type *>(&(coroutine.promise())));
        coroutine.destroy();
      }
      else
      {
        _coroutine = coroutine;
      }
    }

    Future(const Future &) = delete;

    Future& operator=(const Future &) = delete;

    Future(Future && other) noexcept
    {
      *this = other;
      other.owning = false;
    }

    Future& operator=(Future && other) noexcept
    {
      *this = other;
      other._owning = false;
    }

    ~Future()
    {
      if (_owning)
      {
        std::destroy_at(&(_result));
      }
      else
      {
        std::destroy_at(&(_coroutine));
      }
    }

    auto get()
    {
      if (_owning)
      {
        return _result.get();
      }
      return _coroutine.promise().get();
    }
  };


  template <typename T = void>
  struct [[nodiscard]] Fork
  {
    struct promise_type;

    coroutine_handle<promise_type> _coroutine;

    struct promise_type : Result<T>
    {
      continuation_handle _continuation;
      coroutine_node _node;

      std::atomic_uint64_t * _n;

      Fork get_return_object() noexcept
      {
        return { coroutine_handle<promise_type>::from_promise(*this) };
      }

      suspend_always initial_suspend() noexcept
      {
        return {};
      }

      auto final_suspend() noexcept
      {
        struct awaitable : suspend_always
        {
          coroutine_handle<> await_suspend(coroutine_handle<promise_type> coroutine) noexcept
          {
            if (std::optional continuation = riften::detail::Forkpool::pop())
            {
              return *continuation;
            }
            std::uint64_t n = coroutine.promise()._n->fetch_sub(1, std::memory_order_acq_rel);
            if (n == 1)
            {
              return coroutine.promise()._continuation.continuation();
            }
            return noop_coroutine();
          }
        };

        return awaitable {};
      }
    };

    struct awaitable
    {
      coroutine_handle<promise_type> _coroutine;
      coroutine_list * _children;
      std::uint64_t _a;
      std::uint64_t * _alpha;

      awaitable(coroutine_handle<promise_type> coroutine) noexcept
        : _coroutine(coroutine)
      {}

      bool await_ready() noexcept
      {
        return false;
      }

      template <typename Promise>
      auto await_suspend(coroutine_handle<Promise> continuation)
      {
        _coroutine.promise()._continuation = continuation;
        _coroutine.promise()._n = &(continuation.promise()._n);

        _children = &(continuation.promise()._children);
        _a = continuation.promise()._alpha;
        _alpha = &(continuation.promise()._alpha);

        std::coroutine_handle<> on_stack_handle = _coroutine;
        riften::detail::Forkpool::push({ continuation, _alpha });
        return on_stack_handle;
      }

      auto await_resume() noexcept
      {
        using result_type = Result<T>;
        static constexpr bool small = sizeof(result_type) <= 2 * sizeof(void*);
        bool stolen = (*_alpha > _a);
        if (stolen)
        {
          _children->insert(_coroutine);
        }
        return Future<promise_type, small>(_coroutine, !stolen);
      }
    };

    auto operator co_await() && noexcept
    {
      return awaitable { _coroutine };
    }
  };


  struct [[nodiscard]] Sync {};


  template <typename T = void>
  struct [[nodiscard]] Join
  {
    struct promise_type;

    coroutine_handle<promise_type> _coroutine;

    ~Join()
    {
      _coroutine.promise()._children.clear();
      _coroutine.destroy();
    }

    struct promise_type : Result<T>
    {
      static constexpr std::uint64_t UMAX = std::numeric_limits<std::uint64_t>::max();

      coroutine_handle<> _continuation;
      coroutine_list _children;

      alignas(riften::hardware_destructive_interference_size) std::atomic_uint64_t _n = UMAX;
      std::uint64_t _alpha = 0;

      Join get_return_object() noexcept
      {
        return { coroutine_handle<promise_type>::from_promise(*this) };
      }

      suspend_always initial_suspend() noexcept
      {
        return {};
      }

      template <typename U>
      decltype(auto) await_transform(U && expr) noexcept
      {
        return std::forward<U>(expr);
      }

      auto await_transform(Sync &&) noexcept
      {
        struct awaitable
        {
          coroutine_handle<promise_type> _coroutine;

          bool await_ready() noexcept
          {
            if (std::uint64_t alpha = _coroutine.promise()._alpha)
            {
              std::uint64_t n = _coroutine.promise()._n.load(std::memory_order_acquire);
              if (n - (UMAX - alpha) == 0)
              {
                return true;
              }
              return false;
            }
            return true;
          }

          coroutine_handle<> await_suspend(coroutine_handle<promise_type> coroutine) noexcept
          {
            std::uint64_t alpha = coroutine.promise()._alpha;
            std::uint64_t n = coroutine.promise()._n.fetch_sub(UMAX - alpha, std::memory_order_acq_rel);
            if (n - (UMAX - alpha) == 0)
            {
              return coroutine;
            }
            return noop_coroutine();
          }

          void await_resume() noexcept
          {
            _coroutine.promise()._alpha = 0;
            _coroutine.promise()._n.store(UMAX, std::memory_order_release);
          }
        };

        return awaitable { coroutine_handle<promise_type>::from_promise(*this) };
      }

      auto final_suspend() noexcept
      {
        struct awaitable : suspend_always
        {
          coroutine_handle<> await_suspend(coroutine_handle<promise_type> coroutine) noexcept
          {
            std::uint64_t alpha = coroutine.promise()._alpha;
            std::uint64_t n = coroutine.promise()._n.fetch_sub(UMAX - alpha, std::memory_order_acq_rel);
            if (n - (UMAX - alpha) == 0)
            {
              return coroutine.promise()._continuation;
            }
            return noop_coroutine();
          }
        };

        return awaitable {};
      }
    };

    auto operator co_await() && noexcept
    {
      struct awaitable
      {
        coroutine_handle<promise_type> _coroutine;

        bool await_ready() noexcept
        {
          return false;
        }

        coroutine_handle<> await_suspend(coroutine_handle<> awaiting_coroutine) noexcept
        {
          _coroutine.promise()._continuation = awaiting_coroutine;
          return _coroutine;
        }

        decltype(auto) await_resume()
        {
          return _coroutine.promise().get();
        }
      };

      return awaitable { _coroutine };
    }
  };


  struct [[nodiscard]] Root
  {
    struct promise_type;

    coroutine_handle<promise_type> _coroutine;

    ~Root()
    {
      _coroutine.destroy();
    }

    struct promise_type : Result<void>
    {
      std::atomic_bool _done = false;

      Root get_return_object() noexcept
      {
        return { coroutine_handle<promise_type>::from_promise(*this) };
      }

      suspend_always initial_suspend() noexcept
      {
        return {};
      }

      auto final_suspend() noexcept
      {
        struct awaitable : suspend_always
        {
          void await_suspend(coroutine_handle<promise_type> coroutine) noexcept
          {
            coroutine.promise()._done.store(true, std::memory_order_release);
            coroutine.promise()._done.notify_one();
          }
        };

        return awaitable {};
      }
    };

    void call() &&
    {
      riften::detail::Forkpool::get();
      std::uint64_t alpha = 0;
      riften::detail::Forkpool::push_root({ _coroutine, &alpha });
      _coroutine.promise()._done.wait(false, std::memory_order_acquire);
    }
  };

}

#endif // __FORKJOIN_HPP__
